///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Implementation on CUDA of the Gaussian mixture model background subtraction from:
//
//  "GPU implementation of GMM for Background Subtraction"
//  Vu Pham, Phong Vo, Vu Thanh Hung, Le Hoai Bac
//  IEEE-RIVF, Nov 2010
//
//  This is the GPU implementation of Zivkovic's algorithm. For more details:
//
//  "Efficient Adaptive Density Estimation per Image Pixel for the Task of Background Subtraction"
//  Z.Zivkovic, F. van der Heijden 
//  Pattern Recognition Letters, vol. 27, no. 7, pages 773-780, 2006.
//
//  and
//
//  "Recursive unsupervised learning of finite mixture models "
//  Z.Zivkovic, F.van der Heijden 
//  IEEE Trans. on Pattern Analysis and Machine Intelligence, vol.26, no.5, pages 651-656, 2004 
//
//
//  Author: Vu Pham, http://phvureseach.wordpress.com
//  University of Science, Ho Chi Minh City, Viet Nam
//  Date: 28-August-2010, Version: 0.9
//
/////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Usage:
//
//  CvFastBgGMMParams* params = cvCreateFastBgGMMParams(FRAME_WIDTH, FRAME_HEIGHT);
//  // modify some parameters in params struct here... 
//  CvFastBgGMM* gmm = cvCreateFastBgGMM(params, frame0, frame1, frame2)
//  // use 3 first frames to initialize
//
//  for each new frame...
//  cvUpdateFastBgGMM(gmm, new_frame);
//  or use the following function to get the processing time:
//  float time = cvUpdateFastBgGMMTimer(gmm, new_frame);
//
//  Finally, call cvReleaseFastBgGMM() to release the memory
//  cvReleaseFastBgGMM(&gmm);
//  delete params;
/////////////////////////////////////////////////////////////////////////////////////////////////////



#include <stdlib.h>
#include <memory.h>

#include <cuda_runtime.h>
#include <cuda.h>

#include <cv.h>
#include <stdio.h>

//////////////////////////////////////////////////////////////////////////////////////////
//
//  CUDAGMM versions:
//  1: __constant__ device memory
//  2: memory coalescing on input image (4-channels input), output image, usedModes array
//  3: SoA: Structure of arrays, pinned memory when transfer input/output frame,
//  	  memory coalescing on SoA (Gaussian data)
//  4: asynchronous execution
//  5: template kernel function <- nearly useless since we 'template' on the BLOCK_SIZE.
//
//  We only have 3, 4, 5 here.
//
//////////////////////////////////////////////////////////////////////////////////////


#ifdef _DEBUG
#define	CUDAGMM_SAFE_CALL(x)	CUDA_SAFE_CALL(x)
#else
#define	CUDAGMM_SAFE_CALL(x)	x
#endif

#include "params.h"

typedef struct CvFastBgGMM
{


	// the 4-channels input image
	IplImage* inputFrame;



	// the output image
	IplImage* h_outputImg;

	// on device memory, just for fast data transfer
	unsigned char* d_outputImg;
	int nOutputImgSize;
	unsigned char* d_inputImg;
	int nInputImgSize;

	// for deallocation only.
	CvFastBgGMMData* internal_data;

	// number of threads per block
	// this will be determine automatically.
	// You SHOULD NOT modify it manually.
	int nThreadsPerBlock;

	// number of blocks per grid
	// this will be determine automatically.
	// You SHOULD NOT modify it manually.
	int nBlocksPerGrid;


	
	// Page-locked host memory
	unsigned char* h_pinnedIn;
	unsigned char* h_pinnedOut;

	cudaStream_t copyStream;




	
	// CUDA streams
	cudaStream_t execStream;

	// double buffer
	unsigned char* d_outputImg2;
	unsigned char* d_inputImg2;


} CvFastBgGMM;

////////////////
//
//  Input:
//	  pGMM - a pointer to an already initialized GMM
//    data - a pointer to an OpenCV 3-channels image
//  Output:
//    pGMM->h_outputImg - an OpenCV single channel image
//    values: 255-foreground, 125-shadow, 0-background
///////////


/*====================================================================================*/
// forward declarations
/*====================================================================================*/

int InitCUDA(CvFastBgGMM* pGMM);
 

extern "C" void cudaUpdateFastBgGMM_Wrapper(unsigned char* data, unsigned char* output, int iGridSize, int iBlockSize, int sharedSize, cudaStream_t stream, 
				int* d_arrImageInfo, CvFastBgGMMData d_GMMData, CvFastBgGMMParams d_GMMParams);


/*====================================================================================*/

/*====================================================================================*/

CvFastBgGMMParams* cvCreateFastBgGMMParams(int width, int height)
{
	CvFastBgGMMParams* pGMMParams = new CvFastBgGMMParams();

	int size = width*height;
	pGMMParams->nWidth = width;
	pGMMParams->nHeight = height;
	pGMMParams->nSize = size;

	pGMMParams->nNBands=3;	//always 3 - not implemented for other values!

	//set parameters
	// K - max number of Gaussians per pixel
	pGMMParams->nM = 4;			
	// Tb - the threshold - n var
	pGMMParams->fTb = 4*4;
	// Tbf - the threshold
	pGMMParams->fTB = 0.9f;//1-cf from the paper 
	// Tgenerate - the threshold
	pGMMParams->fTg = 3.0f*3.0f;//update the mode or generate new
	pGMMParams->fSigma= 11.0f;//sigma for the new mode
	// alpha - the learning factor
	pGMMParams->fAlphaT=0.001f;
	// complexity reduction prior constant
	pGMMParams->fCT=0.05f;

	//shadow
	// Shadow detection
	pGMMParams->bShadowDetection = 1;//turn on
	pGMMParams->fTau = 0.5f;// Tau - shadow threshold

	pGMMParams->bRemoveForeground = 0;
	return pGMMParams;
}
/*====================================================================================*/

/*====================================================================================*/

template <bool toPinned>
void copyImageData(IplImage* h_img, unsigned char* d_pinnedMem, int channels)
{
	if(h_img->widthStep == channels*h_img->width)
	{
		memcpy(
			toPinned ? (d_pinnedMem) : (unsigned char*)(h_img->imageData), 
			toPinned ? (unsigned char*)(h_img->imageData) : (d_pinnedMem),
			h_img->widthStep*h_img->height);
	}
	else
	{
		unsigned char* d_curData = d_pinnedMem;
		if(toPinned)
		{
			for(int i = 0; i < h_img->height; ++i)
			{
				memcpy(
					d_curData,
					&CV_IMAGE_ELEM(h_img, unsigned char, i, 0),
					channels*h_img->width);
				d_curData += (channels*h_img->width);
			}
		}
		else
		{
			for(int i = 0; i < h_img->height; ++i)
			{
				memcpy(
					&CV_IMAGE_ELEM(h_img, unsigned char, i, 0),
					d_curData,
					channels*h_img->width);
				d_curData += (channels*h_img->width);
			}
		}
	}
}

/*====================================================================================*/

/*====================================================================================*/

CvFastBgGMM* cvCreateFastBgGMM(CvFastBgGMMParams* pGMMParams, IplImage* frame0,CvFastBgGMMData* d_GMMData,int *d_arrImageInfo,CvFastBgGMMParams *d_GMMParams)
{
	CvFastBgGMM* h_pGMMRet = new CvFastBgGMM();

	if(InitCUDA(h_pGMMRet) < 0)
	{
		delete h_pGMMRet;
		return NULL;
	}

	CvFastBgGMMData* h_pGMMData = new CvFastBgGMMData();

	// allocate device global memory
	int iElemCount = pGMMParams->nSize * pGMMParams->nM * sizeof(float);
	int iSizeCount = pGMMParams->nSize * sizeof(int);

	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMData->ucGaussian), 4*iElemCount));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMData->rWeight), iElemCount));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMData->rnUsedModes), iSizeCount));
	CUDAGMM_SAFE_CALL(cudaMemset(h_pGMMData->rnUsedModes, 0, iSizeCount));
	 
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)d_GMMData, sizeof(CvFastBgGMMData)));
	CUDAGMM_SAFE_CALL(cudaMemcpy(d_GMMData, h_pGMMData, sizeof(CvFastBgGMMData), cudaMemcpyHostToDevice));

	h_pGMMRet->internal_data = h_pGMMData;

	// we will use 4-channels image as input data!
	h_pGMMRet->inputFrame = cvCreateImage(cvSize(pGMMParams->nWidth, pGMMParams->nHeight), IPL_DEPTH_8U, 4);
	h_pGMMRet->nInputImgSize = 4 * frame0->width *  pGMMParams->nHeight;
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_inputImg), h_pGMMRet->nInputImgSize));
	h_pGMMRet->h_outputImg = cvCreateImage(cvSize(pGMMParams->nWidth, pGMMParams->nHeight), IPL_DEPTH_8U, 1);
	h_pGMMRet->nOutputImgSize = pGMMParams->nWidth *  pGMMParams->nHeight;
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_outputImg), h_pGMMRet->nOutputImgSize));
	
	// d_arrImageInfo constant (device mem.)
	int inpPixelCnt = pGMMParams->nWidth * pGMMParams->nHeight;

	// number of pixels per thread must be 4k, i.e. 4, 8, 12, 16, 20...
	int iPixelsPerThread = (int)ceil(inpPixelCnt *1.0 / (h_pGMMRet->nBlocksPerGrid * h_pGMMRet->nThreadsPerBlock));
	iPixelsPerThread = 4*(int)ceil(iPixelsPerThread/4.0f);
	h_pGMMRet->nBlocksPerGrid = (int)ceil(inpPixelCnt*1.0 / ((h_pGMMRet->nThreadsPerBlock) * iPixelsPerThread));

	printf("%d pixels/thread, %d threads/block, %d blocks\r\n", 
		iPixelsPerThread, h_pGMMRet->nThreadsPerBlock, h_pGMMRet->nBlocksPerGrid);

	int arrImgInfo[ImageInfoCount] = {	inpPixelCnt, iPixelsPerThread };
	
	CUDAGMM_SAFE_CALL(cudaMalloc((void **)d_arrImageInfo, ImageInfoCount*sizeof(int)));
	CUDAGMM_SAFE_CALL(cudaMemcpy(d_arrImageInfo, arrImgInfo, ImageInfoCount*sizeof(int), cudaMemcpyHostToDevice));

	CUDAGMM_SAFE_CALL(cudaStreamCreate(&(h_pGMMRet->copyStream)));
	CUDAGMM_SAFE_CALL(cudaStreamCreate(&(h_pGMMRet->execStream)));
	CUDAGMM_SAFE_CALL(cudaHostAlloc((void**)&(h_pGMMRet->h_pinnedIn), h_pGMMRet->nInputImgSize, cudaHostAllocMapped));
	CUDAGMM_SAFE_CALL(cudaHostAlloc((void**)&(h_pGMMRet->h_pinnedOut), h_pGMMRet->nOutputImgSize, cudaHostAllocMapped));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_inputImg2), h_pGMMRet->nInputImgSize));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_outputImg2), h_pGMMRet->nOutputImgSize));

	// copy the algorithm parameters to Constant memory
	pGMMParams->fPrune = -(pGMMParams->fAlphaT) * (pGMMParams->fCT);
	
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)d_GMMParams, sizeof(CvFastBgGMMParams)));
	CUDAGMM_SAFE_CALL(cudaMemcpy(d_GMMParams, pGMMParams, sizeof(CvFastBgGMMParams), cudaMemcpyHostToDevice));

	// setup the initial state for asynchronous execution
	cvCvtColor(frame0, h_pGMMRet->inputFrame, CV_BGR2BGRA);
	copyImageData<true>(h_pGMMRet->inputFrame, h_pGMMRet->h_pinnedIn, 4);
	CUDAGMM_SAFE_CALL(cudaMemcpy(h_pGMMRet->d_inputImg2, h_pGMMRet->h_pinnedIn, h_pGMMRet->nInputImgSize, cudaMemcpyHostToDevice));
	cudaUpdateFastBgGMM_Wrapper
		( h_pGMMRet->d_inputImg2, h_pGMMRet->d_outputImg2,(h_pGMMRet->nBlocksPerGrid),(h_pGMMRet->nThreadsPerBlock),4,h_pGMMRet->execStream, d_arrImageInfo, *d_GMMData, *d_GMMParams);

	CUDAGMM_SAFE_CALL(cudaMemcpyAsync(h_pGMMRet->d_inputImg, h_pGMMRet->h_pinnedIn, h_pGMMRet->nInputImgSize, 
		cudaMemcpyHostToDevice, h_pGMMRet->copyStream));
	CUDAGMM_SAFE_CALL(cudaStreamSynchronize(h_pGMMRet->execStream));
	CUDAGMM_SAFE_CALL(cudaMemcpy(h_pGMMRet->h_pinnedOut, h_pGMMRet->d_outputImg2, h_pGMMRet->nOutputImgSize, cudaMemcpyDeviceToHost));

	CUDAGMM_SAFE_CALL(cudaStreamSynchronize(h_pGMMRet->copyStream));
	cudaUpdateFastBgGMM_Wrapper
		( h_pGMMRet->d_inputImg, h_pGMMRet->d_outputImg,(h_pGMMRet->nBlocksPerGrid), (h_pGMMRet->nThreadsPerBlock), 4, h_pGMMRet->execStream,d_arrImageInfo, *d_GMMData, *d_GMMParams );

	CUDAGMM_SAFE_CALL(cudaMemcpyAsync(h_pGMMRet->d_inputImg2, h_pGMMRet->h_pinnedIn, h_pGMMRet->nInputImgSize, 
		cudaMemcpyHostToDevice, h_pGMMRet->copyStream));

	return h_pGMMRet;
}

/*====================================================================================*/

/*====================================================================================*/

 void cvReleaseFastBgGMM(CvFastBgGMM** h_ppGMM)
{
	CvFastBgGMM* h_pGMM = *h_ppGMM;

	cvReleaseImage(&(h_pGMM->h_outputImg));
	cvReleaseImage(&(h_pGMM->inputFrame));
	CUDAGMM_SAFE_CALL( cudaStreamSynchronize(h_pGMM->copyStream));
	CUDAGMM_SAFE_CALL( cudaStreamSynchronize(h_pGMM->execStream));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMM->d_inputImg));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMM->d_outputImg));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMM->d_inputImg2));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMM->d_outputImg2));
	CUDAGMM_SAFE_CALL( cudaFreeHost(h_pGMM->h_pinnedIn));
	CUDAGMM_SAFE_CALL( cudaFreeHost(h_pGMM->h_pinnedOut));
	CUDAGMM_SAFE_CALL( cudaStreamDestroy(h_pGMM->copyStream));
	CUDAGMM_SAFE_CALL( cudaStreamDestroy(h_pGMM->execStream));

	CvFastBgGMMData* h_pGMMData = h_pGMM->internal_data;
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMMData->ucGaussian));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMMData->rWeight));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMMData->rnUsedModes));
	
	delete h_pGMM->internal_data;
	delete h_pGMM;
	(*h_ppGMM) = 0;
}

/*====================================================================================*/

/*====================================================================================*/

void cvUpdateFastBgGMM(CvFastBgGMM* pGMM, IplImage* inputImg,int *d_arrImageInfo,CvFastBgGMMData  *d_GMMData,CvFastBgGMMParams *d_GMMParams)
{
	cvCvtColor(inputImg, pGMM->inputFrame, CV_BGR2BGRA);
	CUDAGMM_SAFE_CALL(cudaStreamSynchronize(pGMM->copyStream));
	copyImageData<true>(pGMM->inputFrame, pGMM->h_pinnedIn, 4);
	copyImageData<false>(pGMM->h_outputImg, pGMM->h_pinnedOut, 1);
	
	CUDAGMM_SAFE_CALL(cudaStreamSynchronize(pGMM->execStream));
	unsigned char* pTmp;
	SWAP(pGMM->d_inputImg, pGMM->d_inputImg2, pTmp);
	SWAP(pGMM->d_outputImg, pGMM->d_outputImg2, pTmp);

	CUDAGMM_SAFE_CALL(cudaMemcpyAsync(pGMM->d_inputImg, pGMM->h_pinnedIn, pGMM->nInputImgSize, cudaMemcpyHostToDevice, pGMM->copyStream));
	CUDAGMM_SAFE_CALL(cudaMemcpyAsync(pGMM->h_pinnedOut, pGMM->d_outputImg, pGMM->nOutputImgSize, cudaMemcpyDeviceToHost, pGMM->copyStream));

	cudaUpdateFastBgGMM_Wrapper( pGMM->d_inputImg2, pGMM->d_outputImg2 , 
		(pGMM->nBlocksPerGrid), (pGMM->nThreadsPerBlock), 4, pGMM->execStream,d_arrImageInfo, *d_GMMData, *d_GMMParams);

#ifdef _DEBUG
	cudaError_t error = cudaGetLastError();
	if(error != cudaSuccess)
	{
		printf("CUDA error: %d: %s\r\n", error, cudaGetErrorString(error));
	}
#endif
}

float cvUpdateFastBgGMMTimer(CvFastBgGMM* pGMM, IplImage* inputImg,int *d_arrImageInfo,CvFastBgGMMData  *d_GMMData,CvFastBgGMMParams *d_GMMParams)
{
	cudaEvent_t start, stop;
	float time = 0;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord( start, 0 );
	
	cvUpdateFastBgGMM(pGMM, inputImg,d_arrImageInfo,d_GMMData,d_GMMParams);

	cudaEventRecord( stop, 0 );
	cudaEventSynchronize( stop );
	cudaEventElapsedTime( &time, start, stop );
	cudaEventDestroy( start );
	cudaEventDestroy( stop );
	return time;
}

/*============================================================================*/
// CUDA-related functions
/*============================================================================*/

int InitCUDA(CvFastBgGMM* pGMM)
{




	int count = 0;
	int i = 0;

	cudaGetDeviceCount(&count);
	if(count == 0)
	{
		fprintf(stderr, "There is no device.\n");
		return -1;
	}

	for(i = 0; i < count; i++)
	{
		cudaDeviceProp prop;
		if(cudaGetDeviceProperties(&prop, i) == cudaSuccess)
		{
			if(prop.major >= 1)
			{
				pGMM->nThreadsPerBlock = prop.maxThreadsPerBlock / 4;

				// temporarily hard-code a little here...
				pGMM->nBlocksPerGrid = 256;
				break;
			}
		}
	}
	if(i == count) {
		fprintf(stderr, "There is no device supporting CUDA.\n");
		return -1;
	}
	cudaSetDevice(i);
	return i;


}





